/*
 * seq2ljpg.c --
 *
 * This routine compresses an angiography file containing multiple images
 * at 512 by 512 by 8 bits to a file containing multiple lossless JPEG
 * images.  The file was originally named pnm2ljpg.c but the name was
 * changed to prevent confusion with the orignal file.  The original
 * copyright notice appears below:
 *
 * This hack was created by Chris Moore at the Brown University Institute
 * for Medical Computing.
 *
 *-------------------------------------------------------------------------
 *
 * This is the main routine for the lossless JPEG encoder.
 *
 * Copyright (c) 1994 Kongji Huang and Brian C. Smith.
 * Cornell University
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF CORNELL
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "jpeg.h"
#include "mcu.h"
#include "proto.h"
#include "pnm2ljpg.h"

#define LITTLE_ENDIAN 0
#define BIG_ENDIAN 1
#define ENDIAN 2
#define NO_ENDIAN 3

/*
 * Global and static variables 
 */
int verbose;		/* If verbose!=0, the verbose flag is on. 	*/
int psvSet[7];		/* The PSV set 					*/
int numSelValue;	/* Number of elements in psvPSV 		*/
long inputFileBytes;	/* Total input file bytes 			*/
long outputFileBytes;	/* Total output file bytes 			*/
long totalHuffSym[7];	/* Total bits of Huffman coded categroy symbols */
long totalAddBits[7];	/* Total additional bits 			*/
static FILE *outFile;	/* Output file pointer 				*/
static FILE *tableFile; /* Stream to write the table offsets. */
static int  table_flag; /* Flag to see if we want the table. */

/*
 *--------------------------------------------------------------
 *
 * WriteJpegData --
 *
 *	This is an interface routine to the JPEG library.  The
 *	library calls this routine to output a block of data.
 *
 * Results:
 *	Number of bytes written.  If this is less than numBytes,
 *	it indeicates an error.
 *
 * Side effects:
 *	The number of bytes output this time is added to
 *	outputFileBytes.
 *
 *--------------------------------------------------------------
 */
int
WriteJpegData  (buffer, numBytes)
    char *buffer;               /* Data to write */
    int numBytes;               /* Number of bytes to write */
{
    outputFileBytes+=numBytes;
    return fwrite(buffer, 1, numBytes, outFile);
}

/*
 *--------------------------------------------------------------
 *
 * ArgParser --
 *
 *	Command line parser.
 *
 * Results:
 *	Command line parameters and options are passed out.
 *
 * Side effects:
 *	Exit on error.
 *
 *--------------------------------------------------------------
 */
static void 
ArgParser(enInfo,argc,argv,optimize,verbose,tableFlag,inFile,outFile)
    CompressInfo *enInfo;
    int argc;
    char **argv;
    int *optimize;
    int *verbose;
    int *tableFlag;
    FILE **inFile, **outFile;
{
    int argn;
    char *arg;
    char *usage="pnm2ljpg [-d -h -v -p# -r# -s{0..7}] -t [InFile [outFile] [tableFile] ]";
    int numOfFiles=0;
    int i,psv;
    int skip = 0;
    long header_skip = 0;
    char *tmpbuffer;

    /*
     * default values
     */
    enInfo->restartInRows=0;
    enInfo->Pt=0;
    *optimize=1;
    *inFile=stdin;
    *outFile=stdout;
    tableFile=stderr;
    *tableFlag = 0;
    *verbose=0;
    numSelValue=7;
    for (i=0;i<numSelValue;i++) {
        psvSet[i]=i+1;
    }

    for (argn = 1; argn < argc; argn++) {
        arg=argv[argn];
        if (*arg != '-') { 

           /*
            * process a file name
            */
           if (numOfFiles==0) {
              if ((*inFile=fopen(arg,"r"))==NULL) {
                 fprintf(stderr,"Can't open %s\n",arg);
                 exit(-1);
              }
           }
           if (numOfFiles==1) {
              if ((*outFile=fopen(arg,"w"))==NULL) {
                 fprintf(stderr,"Can't open %s\n",arg);
                 exit(-1);
              }
           }
           if (numOfFiles==2) {
              if ((tableFile=fopen(arg,"w"))==NULL) {
                 fprintf(stderr,"Can't open %s\n",arg);
                 exit(-1);
              }
              *tableFlag=1;
           }
           if (numOfFiles>2) {
              fprintf(stderr,"%s\n",usage);
              exit(-1);
           }
           numOfFiles++;
        }
        else {

           /*
            * precess a option
            */
           arg++;
           switch (*arg) {
             case 'd': 
                       /*
                        * default Huffman table flag
                        */
                       *optimize=0; 
                       break;
             case 'h': 
                       /*
                        * help flag
                        */
                       fprintf(stderr,"Encode a 512 by 512 motion image into ");
                       fprintf(stderr,"a lossless JPEG archive.\n");
                       fprintf(stderr,"Usage:\n");
                       fprintf(stderr,"%s\n",usage);
                       fprintf(stderr,"default input\tstdin\n");
                       fprintf(stderr,"default output\tstdout\n");
                       fprintf(stderr,"default table output\tstderr\n");
                       fprintf(stderr,"-d\t\tuse default Huffman table\n");
                       fprintf(stderr,"-h\t\thelp\n");
                       fprintf(stderr,"-p#\t\t# is the point tranform ");
                       fprintf(stderr,"parameter, a positive integer\n");
                       fprintf(stderr,"\t\tless than ");
                       fprintf(stderr,"sample precision.\n");
                       fprintf(stderr,"-r#\t\trestart at every # rows\n");
                       fprintf(stderr,"-s <nnn> skip nnn bytes of header\n");
                       fprintf(stderr,"-S{1..7}\t{1..7} is a set of predictor");
                       fprintf(stderr," selection values, e.g. -s146.\n");
                       fprintf(stderr,"-v\t\tverbose\n");
		       fprintf(stderr,"-t\t\tCreate a table of offsets to each frame\n");
                       exit(1); 
                       break;

             case 's':

                      /*
                       * header skip.
                       */

                       skip = 1;
                       argn++;
                       arg= argv[argn];
                       header_skip = atoi( arg );
                       break;

             case 'p':
                       /*
                        * point transform parameter
                        */
                       enInfo->Pt=*(++arg)-'0';
                       break;
             case 'S':
                       /*
                        * predictor selection value set
                        */
                       psv=*(++arg)-'0';
                       numSelValue=0;
                       while ((psv>0) && (psv<8)) {
                             psvSet[numSelValue++]=psv;
                             psv=*(++arg)-'0';
                       }
                       break;
             case 'r':
                       /*
                        * restart interval
                        */
                       enInfo->restartInRows=atoi(++arg);
                       break;
             case 'v':
                       /*
                        * verbose flag
                        */
                       *verbose=1;
                       break;
             case 't':
                       /*
                        * verbose flag
                        */
                       *tableFlag=1;
                       break;
             default: 
                       fprintf(stderr,"%s\n",usage);
                       exit(-1);
           }
        }
    }


    if( skip > 0 ) {

        if( (tmpbuffer = (char *)malloc( header_skip )) == NULL ) {

           fprintf(stderr, " Cannot malloc space to skip the header.\n");
           exit( -1 );

        }

        /*
         * read in and throw away the header.
         */

			if(*verbose==1) {

				fprintf(stderr, "Skipping %d byte header in infile.\n", header_skip );

			}
         fread( tmpbuffer, 1, header_skip, *inFile );
         free( tmpbuffer );

     }
}

/*
 *--------------------------------------------------------------
 *
 * StatOutput --
 *
 *      Display statistics data on screen.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Compression statistics data are output to stderr.
 *
 *--------------------------------------------------------------
 */
void
StatOutput(enInfo)
    CompressInfo enInfo;
{
    int i,tbl;

    /* 
     * The total bytes of category symbols and additional bits.
     */
    fprintf(stderr,"PSV HuffSymSum AddBitsSum\n");
    for (i=0;i<numSelValue;i++) {
        tbl=psvSet[i]-1;
        fprintf(stderr,"%3d %10ld %10ld\n",
               psvSet[i],totalHuffSym[tbl]/8,totalAddBits[tbl]/8);
    }

    /* 
     * The PSV selected by our PSV selector.
     * The original and compressed image file size.
     * The compression ration (original/compressed).
     */
    fprintf(stderr,"Selected PSV\t\t: %d\n",enInfo.Ss);
    fprintf(stderr,"Original file size\t: %ld bytes\n",inputFileBytes); 
    fprintf(stderr,"Compressed file size\t: %ld bytes\n",outputFileBytes); 
    fprintf(stderr,"Compression ratio\t:");
    fprintf(stderr,"%5.2f\n",(float)inputFileBytes/(float)outputFileBytes);
}

int checkEndian()
{

	union {
		
		int asInt;
		unsigned char asByte[2];

	} testValue;

	testValue.asInt = 1;
	if ( testValue.asByte[0] == '\1') {

		return( LITTLE_ENDIAN );

	} else {

		return( BIG_ENDIAN );

	}

}

/*
 * The emitBytes() routine is used to write bytes to the output stream.
 * Output is always Little Endian, no matter what CPU it is run on.
 * It is passed a pointer to the data to be written, along with the
 * length of the data field to write, and a flag indicating whether an
 * adjustment for "Endian"ism needs to be made.  If this flag is set to
 * NO_ENDIAN, or if the code is run on a LittleEndian machine, then the
 * bytes are written out directly.  If, however, the flag is set to ENDIAN
 * and this is NOT a LittleEndian machine, then the bytes are written 
 * in reverse order.
 */

void emitBytes( value, length, thisEndian, adjustEndian )
char *value; 
int length; 
int thisEndian; 
int adjustEndian;
{

	int i;
	if((thisEndian == LITTLE_ENDIAN) || (adjustEndian == NO_ENDIAN )) {

		fwrite( value, length, 1, tableFile );

	} else {

		for( i=length-1; i >= 0; i-- ) {

			fwrite( value + i, 1, 1, tableFile );

		}

	}

}

int
main(argc, argv)
    int argc;
    char **argv;
{
    CompressInfo enInfo;
    CompressInfo enInfo_template;
    JpegComponentInfo *compPtr;
    int i;
    int endian;
    long offset=0;
    FILE *inFile; 
    int optimize;
	 static int loop = 1;

    /*
     * Check the endian-ness of this machine.  We want to always
     * write the table of offsets in little endian mode. So we
     * need to understand what we are so the table gets written 
     * correctly.
     */

    endian = checkEndian();

    /*
     * Process command line parameters.
     */



    MEMSET(&enInfo, 0, sizeof(enInfo));
    MEMSET(&enInfo_template, 0, sizeof(enInfo));
    ArgParser(&enInfo,argc,argv,&optimize,&verbose,&table_flag,&inFile,&outFile);
    if (verbose) {
       fprintf(stderr,"Point transform parameter is %d.\n",enInfo.Pt);
       fprintf(stderr,"Restart every %d row(s).\n",enInfo.restartInRows);
    }

	 memcpy( &enInfo_template, &enInfo, sizeof( CompressInfo ));
    /* 
     * Load the image and get ready for encoding.
     */
    while( loop == 1 ) {

	     memcpy( &enInfo, &enInfo_template, sizeof( CompressInfo ));
        if( PmRead(&enInfo,inFile) == -1 ) {

        		 FreeArray2D(mcuTable);
    			 fclose(inFile);
    			 fclose(outFile);
				 exit(1);

        }
        HuffEncoderInit (&enInfo);
    
        /* 
         * Get the Huffman tables ready. Select the best PSV to
         * be used in encoding pass.
         */
        if (optimize) {
    
           /*
            * optimal encoding
            */

           if (verbose) {
              fprintf(stderr,"Use optimal Huffman table.\n");
           }
           
           /*
            * Assign a Huffman table to be used by a component.
            */
           for (i = 0; i < enInfo.compsInScan; i++) {
               compPtr = enInfo.curCompInfo[i];
               compPtr->dcTblNo=i;
           }
    
           /*
            * Count the times each difference category occurs. 
            * Construct the optimal Huffman table, then select
            * the best PSV using optimal Huffman table.
            */
           HuffOptimize (&enInfo);
            }
            else {
    
           /* 
            * non-optimal encoding
            */
           if (verbose) {
              fprintf(stderr,"Use default huffman table.\n");
           }
    
           /*
            * Assign a Huffman table to be used by a component.
            * In non-optimal encoding, all components share one 
            * stardard Huffman table. 
            */
           for (i = 0; i < enInfo.compsInScan; i++) {
               compPtr = enInfo.curCompInfo[i];
               compPtr->dcTblNo=0;
           }
    
           /*
            * Load and prepare the standard Huffman table.
            */ 
           LoadStdHuffTables(&enInfo);
           for (i = 0; i < enInfo.compsInScan; i++) {
               compPtr = enInfo.curCompInfo[i];
               FixHuffTbl ((enInfo.dcHuffTblPtrs)[compPtr->dcTblNo]);
           }
    
           /*
            * Select the best PSV using standard Huffman table.
            */
           if (numSelValue>1) {
              StdPickSelValue(&enInfo);
              }
           else {
              enInfo.Ss=psvSet[0];
           }
        }
    
        /*
         * Write the frame and scan headers. Encode the image.
         * Clean up everything. Output statistics if verbose
         * flag is on.
         */
        WriteFileHeader (&enInfo); 
        WriteScanHeader (&enInfo);
        HuffEncode (&enInfo);
        FreeArray2D(mcuTable);
        HuffEncoderTerm ();
        WriteFileTrailer (&enInfo);
        FlushBytes ();
        fflush(outFile);
        if (ferror(outFile)) {
	            fprintf (stderr, "Output file write error\n");
        }
        if (verbose) {
       	    StatOutput(enInfo);
        }
	 FreeMcuTable();

	 /*
	  * Print out the table.
	  */

	if( table_flag == 1 ) {

	  emitBytes((char *)&offset, sizeof(long), endian, LITTLE_ENDIAN, tableFile);
	  offset += outputFileBytes;
	  outputFileBytes = 0;

	}
    }
    fclose(inFile);
    fclose(outFile);
}
